const grpc = require('grpc')
const protoLoader = require('@grpc/proto-loader')

class Client {
  constructor(options) {
    const opts = Client.sanitize(options)

    if (!opts.service || !opts.namespace) {
      throw new Error('Please define namespace and service.')
    }

    if (!opts.protoFile) {
      throw new Error('Please provide the proto file path.')
    }

    this.server = opts.server
    this.port = opts.port
    this.namespace = opts.namespace
    this.service = opts.service

    try {
      const packageDefinition = protoLoader.loadSync(opts.protoFile, { keepCase: true })
      const packageObject = grpc.loadPackageDefinition(packageDefinition)

      if (!packageObject[this.namespace]) {
        throw new Error(`Namespace ${this.namespace} not found.`)
      }

      if (!packageObject[this.namespace][this.service]) {
        throw new Error(`Service ${this.service} not found.`)
      }

      this.client = new packageObject[this.namespace][this.service](
        this.getConnectionString(),
        grpc.credentials.createInsecure()
      )
    } catch (err) {
      if (err.code === 'ENOENT') {
        throw new Error('Invalid proto file.')
      }

      throw err
    }
  }

  close() {
    this.client.close()
  }

  invoke(rpcName, data) {
    return new Promise((resolve, reject) => {
      if (typeof this.client[rpcName] !== 'function') {
        return reject(new Error(`No function ${rpcName} defined in the service`))
      }

      this.client[rpcName](data, (err, result) => {
        if (err) {
          return reject(err)
        }

        return resolve(result)
      })
    })
  }

  getConnectionString() {
    return `${this.server}:${this.port}`
  }
}

Client.sanitize = function(options) {
  return {
    server: 'localhost',
    port: 50051,
    service: null,
    namespace: null,
    ...options
  }
}

module.exports = Client
