'use strict'

const protoLoader = require('@grpc/proto-loader')
const test = require('ava')
const grpc = require('grpc')
const path = require('path')
const util = require('util')

const Client = require('../lib/client')
const validOptions = require('./fixtures/valid-options')

function createGrpcServer() {
  // Create a server which responds to the service
  const packageDefinition = protoLoader.loadSync(path.join(__dirname, 'fixtures/helloworld.proto'), { keepCase: true })
  const server = new grpc.Server()

  server.addService(packageDefinition['helloworld.Greeter'], {
    SayHello(call, cb) {
      if (!call.request.name) {
        return cb(new Error('Missing name'))
      }

      return cb(null, { message: `Hello ${call.request.name}!` })
    }
  })

  const port = server.bind('127.0.0.1:0', grpc.ServerCredentials.createInsecure())
  server.start()

  return { port, close: util.promisify(server.tryShutdown.bind(server)) }
}

// Sanitize
test('.sanitize - set default options if key is missing', t => {
  const output = Client.sanitize({})

  t.is('localhost', output.server)
  t.is(50051, output.port)
})

test('.sanitize - not set default options if key is present', t => {
  const output = Client.sanitize({ server: 'anotherserver' })

  t.is('anotherserver', output.server)
  t.is(50051, output.port)
})

test('.constructor - should pass if all options are passed', t => {
  t.notThrows(() => new Client(validOptions))
})

test('.constructor - should throw if no options are passed', t => {
  const tests = [undefined, null, {}, { server: 'asd', port: 1234 }]

  tests.forEach(testData => {
    t.throws(() => new Client(testData), 'Please define namespace and service.')
  })
})

test('.constructor - should throw if no proto file is passed', t => {
  const baseData = {
    service: 'MyService',
    namespace: 'namespace'
  }

  const testData = [undefined, null]

  testData.forEach(td => {
    baseData.protoFile = td
    t.throws(() => new Client(baseData), 'Please provide the proto file path.')
  })

  t.throws(() => {
    baseData.protoFile = 'an/invalid/path'
    return new Client(baseData)
  }, 'Invalid proto file.')
})

test('.constructor - should throw if namespace is not defined in proto file', t => {
  t.throws(() => new Client({ ...validOptions, namespace: 'badnamespace' }), 'Namespace badnamespace not found.')
})

test('.constructor - should throw if service is not defined in proto file', t => {
  t.throws(() => new Client({ ...validOptions, service: 'BadService' }), 'Service BadService not found.')
})

test('.getConnectionString - should get correct connection string', t => {
  const c = new Client(validOptions)

  t.is('myserver:1234', c.getConnectionString())
})

test('.invoke - should handle successful method invocations', async t => {
  const { port, close } = createGrpcServer()
  const c = new Client({ ...validOptions, server: '127.0.0.1', port })

  t.deepEqual({ message: 'Hello World!' }, await c.invoke('SayHello', { name: 'World' }))
  await close()
})

test('.invoke - should handle non existing method invocations', async t => {
  const { port, close } = createGrpcServer()
  const c = new Client({ ...validOptions, server: '127.0.0.1', port })

  await t.throwsAsync(() => c.invoke('SaySomething', { name: 'World' }), {
    message: 'No function SaySomething defined in the service'
  })
  await close()
})

test('.invoke - should handle non existing failed invocations', async t => {
  const { port, close } = createGrpcServer()
  const c = new Client({ ...validOptions, server: '127.0.0.1', port })

  await t.throwsAsync(() => c.invoke('SayHello', { name: '' }), {
    message: '2 UNKNOWN: Missing name'
  })
  await close()
})
